/**
 *  jQuery Font Awesome UI - Icon Browser.
 *
 * A versatile icon picker leveraging Font Awesome icons and jQuery,
 * specifically designed for the Drupal Font Awesome UI module.
 *
 * This plugin has been comprehensively rewritten and optimized
 * to fully support the Font Awesome 6 library, ensuring a seamless
 * and efficient user experience.
 *
 *  https://www.drupal.org/project/fontawesome_ui
 *
 *  Made by Mahyar Sabeti:
 *
 * {@link https://www.drupal.org/u/mahyarsbt}
 *
 *  Licensed under MIT License.
 */
(function ($) {
  /**
   * Initializes the FontAwesome icon browser.
   *
   * This plugin creates a modal browser for selecting FontAwesome icons.
   *
   * @param {Object} options
   *   Configuration options for the icon browser.
   * @param {Object} options.source
   *   The source of icons to populate the browser.
   * @param {Object} options.messages
   *   Customizable browser UI messages, such as labels and placeholders.
   * @param {string} [options.idSuffix]
   *   Suffix to uniquely identify modal elements.
   * @param {boolean} [options.allowEmpty=false]
   *   Whether to allow empty icon selection.
   * @param {boolean} [options.autoLoad=false]
   *   Whether the browser should auto-open upon initialization.
   * @param {string} [options.selectedIcon]
   *   The icon class pre-selected when the browser loads.
   *
   * @return {jQuery}
   *   The jQuery collection this plugin was initialized on.
   */
  $.fn.fontAwesomeIconBrowser = function (options) {
    // Default configuration options.
    const defaults = {
      source: {},
      messages: {
        icon_browser: 'Icon Browser',
        close_label: 'Close',
        search_placeholder: 'Search ...',
        insert_label: 'Insert'
      },
      idSuffix: '',
      allowEmpty: false,
      autoLoad: false,
      selectedIcon: '',
      clickInsert: false,
      showCount: true,
      itemsPerPage: 36,
      version: '6.6.0',
      license: 'free'
    };

    // Merge user options with defaults.
    const settings = $.extend(true, {}, defaults, options);

    // Initialize the browser instance.
    const browser = new FontAwesomeIconBrowser(settings);

    // Use `this.getStyleAlias` from the `FontAwesomeIconBrowser` instance.
    if (settings.selectedIcon) {
      settings.selectedIcon = browser.getStyleAlias(settings.selectedIcon);
    }

    // Automatically open the modal if `autoLoad` is enabled.
    if (settings.autoLoad) {
      $(document).ready(function () {
        browser.openModal();
      });
    }

    // Initialize the plugin for each matched element.
    return this.each(function () {
      const $input = $(this);
      $input.css({
        visibility: 'hidden',
        position: 'relative',
        zIndex: '-1',
        display: 'inline-block',
        height: '40px',
        width: '102px',
        padding: '0px',
        margin: '0px -102px 0px 0px',
        border: '0px none',
        verticalAlign: 'top',
        float: 'none'
      });

      // Create the icon selector.
      const $iconSelector = $(`
        <div class="icons-selector faui-darkgrey" style="position: relative;" data-faui-origin="${$input.attr('id')}">
          <div class="selector" data-faui-origin="${$input.attr('id')}">
            <span class="selected-icon"><i class="${settings.selectedIcon}"></i></span>
            <span class="selector-button">
              <i class="faui-icon-down-dir"></i>
            </span>
          </div>
        </div>
      `);

      // Add the selector to the DOM and bind event listeners.
      $input.before($iconSelector);
      $iconSelector.find('.selector').click(function () {
        browser.openModal();
      });

      browser.$input = $input;
      browser.$iconSelector = $iconSelector;
    });
  };

  /**
   * FontAwesomeIconBrowser class.
   *
   * Handles the core logic for the FontAwesome icon browser.
   */
  class FontAwesomeIconBrowser {
    /**
     * Constructor for FontAwesomeIconBrowser.
     *
     * @param {Object} settings
     *   Configuration settings for the browser.
     */
    constructor(settings) {
      this.settings = settings;
      this.currentPage = 1;
      this.iconsPerPage = settings.itemsPerPage;
      this.totalPages = 1;
      this.source = settings.source;
      this.allowEmpty = settings.allowEmpty;
      this.modalElement = null;
      this.sidebarTabs = null;
      this.sidebarTabsGroupStyle = null;
      this.sidebarTabsGroupCategories = null;
      this.iconContainer = null;
      this.footerElement = null;
      this.searchInput = null;
      this.searchQuery = '';
      this.activeCategory = 'all_categories';
      this.activeStyle = 'all';
      this.activeFamily = 'all';
      this.activeType = 'all';
      this.hasProIcons = this.checkForProIcons();
      this.hasSharpIcons = this.checkForSharpIcons();
      this.version = settings.version;
      this.license = settings.license;
      this.init();
    }

    // Define style aliases as a static property for reuse across methods.
    static styleAliases = new Map([
      ['fa-sharp fa-solid', 'fass'],
      ['fa-sharp fa-regular', 'fasr'],
      ['fa-sharp fa-light', 'fasl'],
      ['fa-sharp fa-thin', 'fast'],
      ['fa-duotone', 'fad'],
      ['fa-duotone fa-solid', 'fad'],
      ['fa-duotone fa-regular', 'fadr'],
      ['fa-duotone fa-light', 'fadl'],
      ['fa-duotone fa-thin', 'fadt'],
      ['fa-sharp fa-duotone', 'fasd'],
      ['fa-sharp-duotone fa-solid', 'fasds'],
      ['fa-sharp-duotone fa-regular', 'fasdr'],
      ['fa-sharp-duotone fa-light', 'fasdl'],
      ['fa-sharp-duotone fa-thin', 'fasdt'],
      ['fa-solid', 'fas'],
      ['fa-regular', 'far'],
      ['fa-light', 'fal'],
      ['fa-thin', 'fat'],
      ['fa-duotone', 'fad'],
      ['fa-brands', 'fab']
    ]);

    /**
     * Initializes the FontAwesome icon browser.
     *
     * This method sets up the modal, event listeners, and visual components.
     */
    init() {
      this.createModal();
      this.attachEventListeners();
      this.renderSidebarTabs();
      this.renderIcons();
      this.toggleElementsVisibility();
    }

    /**
     * Retrieves all unique styles from the icon source.
     *
     * Iterates through the source data to collect unique icon styles. Uses a
     * Set to ensure each style appears only once for efficiency.
     *
     * @return {Array}
     *   An array of unique styles found in the source data.
     */
    getAllStyles() {
      const styles = new Set();

      // Add each icon's style to the Set.
      Object.values(this.source).forEach(category => {
        Object.values(category).forEach(iconData => {
          iconData.icons.forEach(icon => {
            styles.add(icon.style);
          });
        });
      });

      // Convert the Set to an array and return it.
      return Array.from(styles);
    }

    /**
     * Resolves the alias for a given class string.
     *
     * Matches the class string to predefined style aliases and returns
     * the alias with any remaining class strings.
     *
     * @param {string} classString
     *   The class string to resolve (e.g., "fa-solid fa-sharp").
     *
     * @return {string}
     *   Combines the resolved alias and unmatched class strings;
     *   returns empty if invalid.
     */
    getStyleAlias(classString) {
      // Handle empty or undefined input gracefully.
      if (!classString) return '';

      // Split the input string into individual classes.
      const classes = classString.split(' ');

      // Variables to track the alias and unmatched classes.
      let styleAlias = '';
      const remainingClasses = [...classes];

      // Sort aliases by specificity (longer keys first for exact matches).
      const sortedAliases = Array.from(FontAwesomeIconBrowser.styleAliases.entries())
        .sort((a, b) => b[0].split(' ').length - a[0].split(' ').length);

      // Search for the most specific alias match.
      for (const [aliasKey, aliasValue] of sortedAliases) {
        const aliasClasses = aliasKey.split(' ');

        // Check if all alias classes exist in the input classes.
        if (aliasClasses.every(cls => remainingClasses.includes(cls))) {
          styleAlias = aliasValue; // Assign the matching alias.

          // Remove matched classes from the remainingClasses array.
          aliasClasses.forEach(cls => {
            const index = remainingClasses.indexOf(cls);
            if (index > -1) {
              remainingClasses.splice(index, 1);
            }
          });

          break; // Exit after finding the most specific match.
        }
      }

      // Combine the resolved alias with any unmatched classes.
      return [styleAlias, ...remainingClasses].filter(Boolean).join(' ');
    }

    /**
     * Creates the modal for the FontAwesome icon browser.
     *
     * Dynamically constructs a modal element with required classes, styles,
     * and elements based on the version and type of FontAwesome. The modal
     * includes headers, sidebars, icon previews, pagination, and a search
     * field. The method also appends the constructed modal element to the DOM.
     */
    createModal() {
      // Extract major and minor versions from the version string.
      const major = this.version.split('.')[0];
      const minor = this.version.split('.')[1];

      // Generate dynamic class names based on the version and type.
      const versionClass = `faui-v${major}-${minor}`;
      const versionMajor = `faui-v${major}`;
      const typeClass = this.license === 'pro' ? 'faui-pro' : 'faui-free';

      // Determine the correct icon classes for close and search actions.
      const closeIconClass = this.version.startsWith('6') ? 'fa-xmark' : 'fa-times';
      const searchIconClass = this.version.startsWith('6')
        ? 'fa-solid fa-magnifying-glass fa-flip-horizontal fa-2x fa-fw'
        : 'fas fa-search';

      // Check for support of extra font families for versions >= 6.7.0 (Pro).
      const supportsDuotone =
        this.version.startsWith('6') &&
        parseFloat(this.version) >= 6.7 &&
        this.license === 'pro';

      // Define the base font families to display in the modal.
      const fontFamilies = [
        { family: 'classic', icon: 'fas fa-icons', label: 'Classic' },
        ...(supportsDuotone ? [{ family: 'duotone', icon: 'fa-duotone fa-icons', label: 'Duotone' }] : []),
        { family: 'sharp', icon: 'fa-sharp fa-solid fa-icons', label: 'Sharp' },
        ...(supportsDuotone ? [{ family: 'sharp-duotone', icon: 'fa-sharp-duotone fa-icons', label: 'Sharp Duotone' }] : []),
        { family: 'brands', icon: 'fab fa-font-awesome-flag', label: 'Brands' },
      ];

      // Define the toggle button for switching between free and pro icons.
      const typeToggle = this.license === 'pro'
        ? `<div class="ais-ToggleRefinement">
            <button class="icons-facet-tab" data-type="free">
              <i class="fa fa-bolt fa-fw fa-2x"></i>
              <span>Free</span>
            </button>
          </div>`
        : '';

      // Construct the modal element dynamically using a template string.
      this.modalElement = $(`
        <div class="faui-modal ${versionClass} ${versionMajor} ${typeClass} ${supportsDuotone ? 'extended-families' : ''}" id="faui-modal${this.settings.idSuffix}" style="display: none;">
          <div class="faui-modal--content">
            <header class="faui-modal--header">
              <div class="faui-modal--header-logo-area">
                <i class="far fa-folder-open"></i>
                <span class="faui-modal--header-logo-title">${this.settings.messages.icon_browser}</span>
              </div>
              <div class="faui-modal--header-close-btn">
                <i class="fas ${closeIconClass}"></i>
              </div>
            </header>
            <div class="faui-modal--body claro-details__wrapper--accordion-item">
              <div class="faui-modal--sidebar">
                <div class="faui-modal--sidebar-tabs"></div>
              </div>
              <div class="faui-modal--icon-preview-wrap list-scrollbar scrollbar-2x">
                <header class="search-sticky">
                  <div class="faui-icon-families">
                    <div class="icons-facets-group-families">
                      ${fontFamilies.map(
                        family => `
                          <button class="icons-facet-tab" data-family="${family.family}">
                            <i class="${family.icon} fa-fw fa-2x"></i>
                            <span>${family.label}</span>
                          </button>`
                      ).join('')}
                    </div>
                    ${typeToggle}
                  </div>
                  <div class="search-sticky__items${supportsDuotone ? '' : ' layout-container'}">
                    <div class="search-sticky__items__inner faui-modal--icon-search">
                      <i class="${searchIconClass}"></i>
                      <input name="" value="" placeholder="${this.settings.messages.search_placeholder}">
                    </div>
                  </div>
                </header>
                <div class="faui-modal--icon-preview-inner">
                  <div class="search-sticky-shadow search-sticky--border"></div>
                  <div class="faui-stats">
                    <div class="icons-results-summary display-flex flex-row flex-wrap">
                      <div class="icons-results-summary-count">
                        <h1 class="h4 text-nowrap">0 Icons</h1>
                      </div>
                      <div class="icons-results-summary-pages">
                        <p class="page-info muted text-nowrap">
                          <span class="sr-only">Currently on&nbsp;</span>
                          <span class="text-capitalize">page</span>
                          <span class="current-page">1</span> of <span class="total-pages">1</span>
                        </p>
                      </div>
                    </div>
                  </div>
                  <div class="faui-modal--icon-preview"></div>
                </div>
              </div>
            </div>
            <footer class="faui-modal--footer">
              <span class="page-info"></span>
              <nav class="pagination pager" role="navigation" aria-labelledby="pagination-heading">
                <ul class="page-buttons pager__items js-pager__items ${!this.settings.clickInsert ? 'pager__items--insert' : ''}"></ul>
              </nav>
              ${this.settings.clickInsert
                ? ''
                : `<button class="faui-insert-icon-button button js-form-submit form-submit" disabled>${this.settings.messages.insert_label}</button>`
              }
            </footer>
          </div>
        </div>
      `);

      // Append the modal to the document body for rendering.
      $('body').append(this.modalElement);

      // Initialize key elements for interaction within the modal.
      this.sidebarTabs = this.modalElement.find('.faui-modal--sidebar-tabs');
      this.iconContainer = this.modalElement.find('.faui-modal--icon-preview');
      this.footerElement = this.modalElement.find('.pagination .page-info');
      this.searchInput = this.modalElement.find('.faui-modal--icon-search input');
      this.insertButton = this.modalElement.find('.faui-insert-icon-button');
      this.statsElement = this.modalElement.find('.faui-stats .icons-results-summary-count h1');
      this.currentPageElement = this.modalElement.find('.faui-stats .current-page');
      this.totalPagesElement = this.modalElement.find('.faui-stats .total-pages');
    }

    /**
     * Attaches event listeners for browser interactions.
     *
     * Binds click, input, and scroll events to DOM elements within the modal.
     */
    attachEventListeners() {
      // Bind close button click to the closeModal method.
      this.modalElement.find('.faui-modal--header-close-btn i')
        .on('click', this.closeModal.bind(this));

      // Attach debounced search input listener for the search field.
      this.searchInput.on('input', this.debounce(this.handleSearch.bind(this), 300));

      // Handle sidebar tab clicks to toggle styles and categories.
      this.sidebarTabs.on('click', '.faui-modal--sidebar-tab-item', (event) => {
        const target = $(event.currentTarget);
        const libraryId = target.data('library-id');
        const libraryStyle = target.data('library-style');
        if (libraryStyle) {
          this.activeStyle = libraryStyle;
          this.sidebarTabs.find('.faui-modal--sidebar-tab-item[data-library-style]')
            .removeClass('style-active');
          target.addClass('style-active');
        } else {
          this.activeCategory = libraryId;
          this.sidebarTabs.find('.faui-modal--sidebar-tab-item[data-library-id]')
            .removeClass('category-active');
          target.addClass('category-active');
        }
        this.currentPage = 1; // Reset to the first page after selection.
        this.renderIcons();
        this.updateSidebarCounts();
      });

      // Manage clicks on individual icons in the icon container.
      this.iconContainer.on('click', '.faui-icon-item', (event) => {
        const target = $(event.currentTarget);
        this.iconContainer.find('.faui-icon-item').removeClass('icon-active');
        target.addClass('icon-active');
        const selectedIconClass = target.data('filter');
        this.settings.selectedIcon = selectedIconClass;

        if (this.settings.clickInsert) {
          this.insertIcon(); // Insert icon directly if clickInsert is true.
        } else {
          this.updateInsertButtonState(); // Enable/disable the insert button.
          this.insertButton.off('click');
          this.insertButton.on('click', this.insertIcon.bind(this));
        }
      });

      // Handle family and type toggles in the icon families section.
      this.modalElement.find('.faui-icon-families').on('click', '.icons-facet-tab', (event) => {
        const target = $(event.currentTarget);
        const family = target.data('family');
        const type = target.data('type');

        if (!target.hasClass('facet-active')) {
          target.addClass('facet-active');
        } else {
          target.removeClass('facet-active');
        }

        if (family) {
          this.activeFamily = this.activeFamily === family ? 'all' : family;
        }
        if (type) {
          this.activeType = this.activeType === type ? 'all' : type;
        }

        this.currentPage = 1; // Reset to the first page after toggle.
        this.renderIcons();
        this.updateSidebarCounts();
      });

      // Add a scroll listener to handle sticky search bar visibility.
      document.addEventListener('DOMContentLoaded', function () {
        const scrollMain = document.querySelector('.faui-modal--icon-preview-wrap');
        const searchSticky = document.querySelector('.search-sticky');
        const searchBorder = document.querySelector('.search-sticky-shadow');

        scrollMain.addEventListener('scroll', function () {
          if (scrollMain.scrollTop >= 50) {
            searchSticky.classList.add('search-sticky--is-sticky');
            searchBorder.classList.remove('search-sticky--border');
          } else {
            searchSticky.classList.remove('search-sticky--is-sticky');
            searchBorder.classList.add('search-sticky--border');
          }
        });
      });
    }

    /**
     * Debounces a function to limit its invocation rate.
     *
     * Ensures the given function is executed only once after the specified
     * delay, even if called repeatedly in quick succession.
     *
     * @param {Function} func
     *   The function to debounce.
     * @param {number} wait
     *   The number of milliseconds to wait before invoking the function.
     *
     * @return {Function}
     *   A debounced version of the provided function.
     */
    debounce(func, wait) {
      let timeout;
      return function (...args) {
        const context = this;
        if (timeout) clearTimeout(timeout);
        timeout = setTimeout(() => {
          timeout = null;
          func.apply(context, args);
        }, wait);
      };
    }

    /**
     * Opens the icon browser modal.
     *
     * Displays the modal and highlights the currently selected icon as active.
     */
    openModal() {
      this.modalElement.show();
      this.setIconActive();
    }

    /**
     * Closes the icon browser modal.
     *
     * Hides the modal element from view.
     */
    closeModal() {
      this.modalElement.hide();
    }

    /**
     * Handles the search input for filtering icons.
     *
     * Updates the search query, resets the current page to 1, and refreshes
     * the displayed icons and sidebar counts.
     *
     * @param {Event} event
     *   The input event triggered by typing in the search field.
     */
    handleSearch(event) {
      this.searchQuery = event.target.value.toLowerCase();
      this.currentPage = 1; // Reset to the first page after a new search.
      this.renderIcons(); // Refresh the icon grid with filtered results.
      this.updateSidebarCounts(); // Update sidebar counts based on the search.
    }

    /**
     * Filters icons based on active filters and search query.
     *
     * Applies filters for style and search terms to the icon list. Icons are
     * retrieved from the sidebar filter results and further narrowed down
     * based on the active style.
     *
     * @return {Array}
     *   A filtered array of icons matching the active filters and query.
     */
    filterIcons() {
      let icons = this.filterIconsBySidebar();

      if (this.activeStyle !== 'all') {
        icons = icons.filter(icon => icon.style === this.activeStyle);
      }

      return icons;
    }

    /**
     * Renders the sidebar tabs for styles and categories.
     *
     * Tabs are dynamically created for styles and categories from the source.
     * If the `showCount` setting is enabled, the count of items is shown next
     * to each tab.
     */
    renderSidebarTabs() {
      // Clear the existing sidebar tabs to ensure fresh rendering.
      this.sidebarTabs.empty();

      // Get the total number of icons for display in the "All styles" section.
      const totalIcons = this.filterIcons().length;

      // Render the "Style" section header.
      this.sidebarTabs.append(`
        <h3 class="h6-muted-text-uppercase">Style</h3>
        <ul class="icons-facets-group-styles"></ul>
      `);

      // Initialize the style group container for dynamic style appending.
      this.sidebarTabsGroupStyle = this.modalElement.find('.icons-facets-group-styles');

      // Add the "All styles" tab as the default active tab.
      this.sidebarTabsGroupStyle.append(`
        <li class="faui-modal--sidebar-tab-item style-active" data-library-style="all">
          <span style="flex: 1 1 auto;">
            <i class="fas fa-grid-2 fa-record-vinyl"></i>All styles
          </span>
          ${this.settings.showCount ? `<span class="icons-facet-count margin-left-sm">${totalIcons}</span>` : ''}
        </li>
      `);

      // Define base style icons and overrides for specific versions.
      const styleIconsBase = {
        'solid': 'circle',
        'regular': this.license === 'pro' ? 'adjust' : 'circle',
        'light': 'circle',
        'thin': 'circle',
        'duotone': 'adjust',
      };
      const styleIconsVersion6Overrides = {
        'regular': this.license === 'pro' ? 'circle-half-stroke' : 'circle',
        'light': 'circle-half-stroke',
        'thin': 'circle-half-stroke',
        'duotone': 'circle-half',
      };
      const styleIcons = this.version.startsWith('6')
        ? { ...styleIconsBase, ...styleIconsVersion6Overrides }
        : styleIconsBase;

      // Render the styles dynamically based on available styles.
      this.getAllStyles().forEach(style => {
        const styleCount = this.filterIcons().filter(icon => icon.style === style).length;

        if (styleCount > 0 && style !== 'brands') {
          // Get the alias for the current style using getStyleAlias.
          const aliasClass = this.getStyleAlias(`fa-${style}`);
          // Capitalize style name for display.
          const styleName = style.charAt(0).toUpperCase() + style.slice(1);
          // Get the icon for the style.
          const styleIcon = styleIcons[style] || 'circle';

          // Append each style as a sidebar tab with the correct icon.
          this.sidebarTabsGroupStyle.append(`
            <li class="faui-modal--sidebar-tab-item" data-library-style="${style}">
              <span style="flex: 1 1 auto;">
                <i class="${aliasClass} fa-${styleIcon}"></i>${styleName}
              </span>
              ${this.settings.showCount ? `<span class="icons-facet-count margin-left-sm">${styleCount}</span>` : ''}
            </li>
          `);
        }
      });

      // Render the "Categories" section header.
      this.sidebarTabs.append(`
        <h3 class="h6-muted-text-uppercase">Categories</h3>
        <div class="fa-ui--refinement-list list-scrollbar">
          <ul class="icons-facets-group-categories"></ul>
        </div>
      `);

      // Set up the categories container to dynamically append categories.
      this.sidebarTabsGroupCategories = this.modalElement.find('.icons-facets-group-categories');

      // Add the "All Categories" tab as the default active tab.
      this.sidebarTabsGroupCategories.append(`
        <li class="faui-modal--sidebar-tab-item category-active" data-library-id="all_categories">
          <span class="icons-facet-affordance position-relative margin-right-sm">
            <i class="fas fa-layer-group fa-fw"></i>
          </span>
          <span style="flex: 1 1 auto;">All Categories</span>
        </li>
      `);

      // Define the base category icons for all versions.
      let categoryIconsBase = {
        'Accessibility' : 'fa-universal-access',
        'Alert' : 'fa-bell',
        'Alphabet' : 'fa-font',
        'Animals' : 'fa-paw',
        'Arrows': this.license === 'pro' ? 'fa-arrow-alt-square-up' : 'fa-caret-square-up',
        'Astronomy' : 'fa-meteor',
        'Audio & Video' : 'fa-play-circle',
        'Automotive' : 'fa-car',
        'Autumn': this.license === 'pro' ? 'fa-pumpkin' : 'fa-leaf',
        'Beverage' : 'fa-coffee',
        'Buildings' : 'fa-building',
        'Business' : 'fa-briefcase',
        'Camping' : 'fa-campground',
        'Charity' : this.license === 'pro' ? 'fa-hands-heart' : 'fa-hand-holding-heart',
        'Chat' : this.license === 'pro' ? 'fa-comments-alt' : 'fa-comments',
        'Chess' : 'fa-chess',
        'Charts + Diagrams' : 'fa-chart-line',
        'Childhood' : 'fa-ice-cream',
        'Clothing' : 'fa-tshirt',
        'Clothing + Fashion' : 'fa-shirt',
        'Code' : 'fa-code',
        'Coding' : 'fa-code',
        'Communication' : 'fa-comment-dots',
        'Computers' : this.license === 'pro' ? 'fa-computer-classic' : 'fa-desktop',
        'Connectivity' : 'fa-signal',
        'Construction' : this.license === 'pro' ? 'fa-digging' : 'fa-paint-roller',
        'Currency' : 'fa-money-bill',
        'Date & Time' : 'fa-clock',
        'Design' : 'fa-palette',
        'Devices + Hardware' : 'fa-computer',
        'Disaster + Crisis' : 'fa-burst',
        'Editing' : 'fa-pencil',
        'Editors' : 'fa-pen',
        'Education' : 'fa-graduation-cap',
        'Emoji' : 'fa-grin-wink',
        'Energy' : 'fa-bolt',
        'Files' : 'fa-file',
        'Film + Video' : 'fa-film',
        'Finance' : this.license === 'pro' ? 'fa-sack-dollar' : 'fa-chart-line',
        'Fitness' : this.license === 'pro' ? 'fa-watch-fitness' : 'fa-hiking',
        'Food' : this.license === 'pro' ? 'fa-cheeseburger' : 'fa-hamburger',
        'Food + Beverage' : 'fa-burger',
        'Fruits & Vegetables' : 'fa-apple-alt',
        'Fruits + Vegetables' : 'fa-apple-whole',
        'Games' : this.license === 'pro' ? 'fa-joystick' : 'fa-gamepad',
        'Gaming' : 'fa-gamepad',
        'Tabletop Gaming' : 'fa-dice-d20',
        'Genders' : 'fa-venus-mars',
        'Halloween' : 'fa-ghost',
        'Hands' : 'fa-hand-paper',
        'Health' : 'fa-heartbeat',
        'Holiday' : this.license === 'pro' ? 'fa-hat-santa' : 'fa-glass-cheers',
        'Holidays' : 'fa-champagne-glasses',
        'Hotel' : 'fa-spa',
        'Household' : 'fa-bath',
        'Humanitarian' : 'fa-parachute-box',
        'Images' : 'fa-images',
        'Interfaces' : 'fa-clipboard',
        'Logistics' : 'fa-shipping-fast',
        'Maps' : 'fa-map',
        'Maritime' : 'fa-anchor',
        'Marketing' : 'fa-bullseye',
        'Mathematics' : 'fa-calculator',
        'Media Playback' : 'fa-forward',
        'Medical' : 'fa-stethoscope',
        'Medical + Health' : 'fa-stethoscope',
        'Money' : 'fa-money-bill-wave',
        'Moving' : 'fa-dolly',
        'Music' : 'fa-music',
        'Music + Audio' : 'fa-music',
        'Nature' : 'fa-tree',
        'Numbers' : 'fa-9',
        'Objects' : 'fa-cut',
        'Payments & Shopping' : 'fa-receipt',
        'Pharmacy' : 'fa-prescription',
        'Photos + Images' : 'fa-image',
        'Political' : 'fa-check-double',
        'Punctuation + Symbols' : 'fa-quote-left',
        'Religion' : 'fa-pray',
        'Science' : 'fa-microscope',
        'Science Fiction' : 'fa-rocket',
        'Security' : this.license === 'pro' ? 'fa-shield' : 'fa-shield-alt',
        'Shapes' : 'fa-shapes',
        'Shopping' : 'fa-shopping-bag',
        'Social' : 'fa-hashtag',
        'Spinners' : 'fa-sync-alt',
        'Sports' : 'fa-running',
        'Sports + Fitness' : 'fa-heart-pulse',
        'Spring' : 'fa-seedling',
        'Status' : 'fa-signal',
        'Summer' : 'fa-sun',
        'Text Formatting' : 'fa-text-height',
        'Time' : 'fa-clock',
        'Toggle' : 'fa-toggle-on',
        'Transportation' : 'fa-truck-plane',
        'Travel' : 'fa-map-marked',
        'Travel + Hotel' : 'fa-plane-departure',
        'Users & People' : 'fa-users',
        'Users + People' : 'fa-children',
        'Vehicles' : this.license === 'pro' ? 'fa-car-bus' : 'fa-bus-alt',
        'Weather' : 'fa-cloud-sun',
        'Winter' :  this.license === 'pro' ? 'fa-snowflakes' : 'fa-snowplow',
        'Writing' : 'fa-keyboard',
        'Sponsors' : this.license === 'pro' ? 'fa-usd-circle' : 'fa-donate',
        'Uncategorized' : 'fa-question-circle',
      };

      // Define version 6 specific overrides for category icons.
      const categoryIconsOverrides = {
        'Arrows': 'fa-square-up-right',
        'Charity' : 'fa-hands-holding-child',
        'Construction' : 'fa-person-digging',
        'Emoji' : 'fa-face-grin-wink',
        'Hands' : 'fa-hand',
        'Logistics' : 'fa-truck-fast',
        'Religion' : 'fa-person-praying',
        'Shopping' : 'fa-basket-shopping',
        'Spinners' : 'fa-arrows-rotate',
        'Sponsors' : 'fa-circle-dollar-to-slot',
        'Uncategorized' : 'fa-circle-question',
      }

      // Merge base and version-specific overrides for categories.
      const categoryIcons = this.version.startsWith('6')
        ? { ...categoryIconsBase, ...categoryIconsOverrides }
        : categoryIconsBase;

      // Render the categories dynamically based on available categories.
      Object.keys(this.source).forEach(category => {
        if (category !== 'Brands') {
          // Count icons for the current category with applied filters.
          const categoryCount = this.filterIcons().filter(icon => {
            return (
              Object.keys(this.source[category]).includes(icon.name) &&
              (this.activeFamily === 'all' || icon.family === this.activeFamily) &&
              (this.activeType === 'all' || icon.type === this.activeType)
            );
          }).length;

          // Get the icon class for the category.
          const iconClass = categoryIcons[category] || 'fa-wand-magic-sparkles';

          // Append the category tab to the sidebar.
          this.sidebarTabsGroupCategories.append(`
            <li class="faui-modal--sidebar-tab-item" data-library-id="${category}">
              <span class="icons-facet-affordance position-relative margin-right-sm">
                <i class="fas ${iconClass} fa-fw"></i>
              </span>
              <span style="flex: 1 1 auto;">${category}</span>
              ${this.settings.showCount ? `<span class="icons-facet-count margin-left-sm">${categoryCount}</span>` : ''}
            </li>
          `);
        }
      });

      // Update counts if `showCount` is disabled.
      if (!this.settings.showCount) {
        this.updateSidebarCounts();
      }
    }

    /**
     * Renders the icons in the main display area.
     *
     * Filters icons based on the active category or search query and renders
     * them in the display area. Also updates pagination and button states.
     */
    renderIcons() {
      // Get the filtered icons for the current category or search query.
      const icons = this.filterIcons();

      // Clear the icon container to prepare for rendering.
      this.iconContainer.empty();

      // Add "None" if empty icons are allowed and on the first page.
      if (this.allowEmpty && this.currentPage === 1) {
        this.iconContainer.append(`
          <div class="faui-icon-item" data-library-id="fa-regular" data-filter=""
            data-library-name="fontAwesome" data-font-style="">
            <div class="faui-icon-item-inner">
              <i class="fa-regular fa- faui-icon-none">&nbsp;</i>
              <div class="faui-icon-item-name" title="None">None</div>
            </div>
          </div>
        `);
      }

      // Calculate the range of icons to display on the current page.
      const start = (this.currentPage - 1) * this.iconsPerPage;
      const end = start + this.iconsPerPage;
      const iconsToShow = icons.slice(start, end);

      // Loop through each icon and render its HTML.
      iconsToShow.forEach(icon => {
        // Convert the alias for the style class (e.g., fab, fas).
        const iconClasses = this.getStyleAlias(icon.class);

        // Extract the icon's name without the "fa-" prefix for display.
        const iconName = icon.name.replace(/^fa-/, '');

        // Append the rendered icon element to the container.
        this.iconContainer.append(`
          <div class="faui-icon-item" data-library-id="${icon.style}"
            data-filter="${iconClasses}" data-library-name="fontAwesome"
            data-font-style="${icon.style}">
            <div class="faui-icon-item-inner">
              <i class="${iconClasses}"></i>
              <div class="faui-icon-item-name" title="${iconName}">
                ${iconName}
              </div>
            </div>
          </div>
        `);
      });

      // Update pagination, button states, and icon statistics.
      this.totalPages = Math.ceil(icons.length / this.iconsPerPage);
      this.updatePagination();
      this.updateInsertButtonState();
      this.updateStats(icons.length);
    }

    /**
     * Updates the selected icon in the UI.
     *
     * Reflects the currently selected icon in the selector and updates the
     * associated input field value.
     */
    updateSelectedIcon() {
      this.$iconSelector.find('.selected-icon').remove();
      this.$iconSelector.find('.selector').prepend(`
        <span class="selected-icon">
          <i class="${this.settings.selectedIcon}"></i>
        </span>
      `);
      this.$input.val(this.settings.selectedIcon).trigger('change');
    }

    /**
     * Sets the active icon in the icon browser.
     *
     * Ensures the selected icon is marked as active within the browser.
     */
    setIconActive() {
      if (this.settings.selectedIcon) {
        this.iconContainer.find('.faui-icon-item').each((index, item) => {
          const $item = $(item);
          const iconClass = $item.data('filter');
          if (iconClass === this.settings.selectedIcon) {
            $item.addClass('icon-active');
          }
        });
      }
      this.updateInsertButtonState();
    }

    /**
     * Updates the state of the insert button.
     *
     * Enables or disables the insert button based on whether an icon is
     * currently selected.
     */
    updateInsertButtonState() {
      const isActive = this.iconContainer
        .find('.faui-icon-item.icon-active').length > 0;
      this.insertButton.prop('disabled', !isActive);
    }

    /**
     * Inserts the selected icon into the input field.
     *
     * Sets the selected icon's class to the input field and closes the modal.
     */
    insertIcon() {
      const activeIcon = this.iconContainer.find('.faui-icon-item.icon-active');
      const iconClass = activeIcon.data('filter');
      this.settings.selectedIcon = iconClass;
      this.updateSelectedIcon();
      this.closeModal();
    }

    /**
     * Checks if the source includes Pro icons.
     *
     * Iterates through the icon source to determine if any Pro icons are
     * available.
     *
     * @return {boolean}
     *   True if Pro icons are available, false otherwise.
     */
    checkForProIcons() {
      for (const category in this.source) {
        for (const iconName in this.source[category]) {
          const iconData = this.source[category][iconName];
          if (iconData.icons.some(icon => icon.type === 'pro')) {
            return true;
          }
        }
      }
      return false;
    }

    /**
     * Checks if the source includes Sharp icons.
     *
     * Iterates through the icon source to determine if any Sharp icons are
     * available.
     *
     * @return {boolean}
     *   True if Sharp icons are available, false otherwise.
     */
    checkForSharpIcons() {
      for (const category in this.source) {
        for (const iconName in this.source[category]) {
          const iconData = this.source[category][iconName];
          if (iconData.icons.some(icon => icon.family === 'sharp')) {
            return true;
          }
        }
      }
      return false;
    }

    /**
     * Toggles visibility of UI elements based on settings and availability.
     *
     * Hides elements related to Pro and Sharp icons if they are not available.
     */
    toggleElementsVisibility() {
      if (!this.hasProIcons) {
        this.modalElement.find('.ais-ToggleRefinement').hide();
      }
      if (!this.hasSharpIcons) {
        this.modalElement
          .find('.icons-facets-group-families button[data-family="sharp"]')
          .hide();
      }
    }

    /**
     * Navigates to the previous page in the icon browser.
     *
     * Decreases the current page index and re-renders icons if the user is
     * not already on the first page.
     *
     * @param {Event} event
     *   The click event triggered by the user.
     */
    prevPage(event) {
      event.preventDefault();
      if (this.currentPage > 1) {
        this.currentPage--;
        this.renderIcons();
      }
    }

    /**
     * Navigates to the next page in the icon browser.
     *
     * Increases the current page index and re-renders icons if the user is
     * not already on the last page.
     *
     * @param {Event} event
     *   The click event triggered by the user.
     */
    nextPage(event) {
      event.preventDefault();
      if (this.currentPage < this.totalPages) {
        this.currentPage++;
        this.renderIcons();
      }
    }

    /**
     * Updates the pagination controls in the UI.
     *
     * Dynamically generates pagination buttons based on the current page
     * and the total number of pages. Includes logic for ellipsis, first
     * and last page links, and next/previous navigation.
     */
    updatePagination() {
      const paginationContainer = this.modalElement
        .find('.pagination .page-buttons');
      paginationContainer.empty();

      const pagesToShow = 6; // Number of pagination links to display.
      const halfPagesToShow = Math.floor(pagesToShow / 2);

      // Determine the range of pages to display.
      let startPage = Math.max(1, this.currentPage - halfPagesToShow);
      let endPage = Math.min(this.totalPages, this.currentPage + halfPagesToShow);

      // Adjust the range if near the start or end of the total pages.
      if (this.currentPage <= halfPagesToShow) {
        endPage = Math.min(this.totalPages, pagesToShow);
      }
      if (this.currentPage + halfPagesToShow >= this.totalPages) {
        startPage = Math.max(1, this.totalPages - pagesToShow + 1);
      }

      this.footerElement.text(`Page ${this.currentPage} of ${this.totalPages}`);

      // Add the "Previous" button.
      paginationContainer.append(`
        <li class="pager__item pager__item--action pager__item--previous">
          <a href="#" title="Go to previous page" rel="prev"
             aria-label="Previous Page" class="prev-page pager__link
             pager__link--action-link">Prev</a>
        </li>
      `);

      // Add the "First page" button and ellipsis if needed.
      if (startPage > 1) {
        paginationContainer.append(`
          <li class="pager__item pager__item--number">
            <a href="#" class="number-page pager__link">1</a>
          </li>
        `);
        if (startPage > 2) {
          paginationContainer.append(`
            <li class="pager__item pager__item--ellipsis" role="presentation">
              …
            </li>
          `);
        }
      }

      // Add buttons for the range of pages.
      for (let i = startPage; i <= endPage; i++) {
        paginationContainer.append(`
          <li class="pager__item ${
          i === this.currentPage ? 'pager__item--active' : ''
        } pager__item--number">
            <a href="#" class="number-page pager__link ${
          i === this.currentPage ? 'is-active' : ''
        }">${i}</a>
          </li>
        `);
      }

      // Add ellipsis and "Last page" button if needed.
      if (endPage < this.totalPages) {
        if (endPage < this.totalPages - 1) {
          paginationContainer.append(`
            <li class="pager__item pager__item--ellipsis" role="presentation">
              …
            </li>
          `);
        }
        paginationContainer.append(`
          <li class="pager__item pager__item--number">
            <a href="#" class="number-page pager__link">
              ${this.totalPages}
            </a>
          </li>
        `);
      }

      // Add the "Next" button.
      paginationContainer.append(`
        <li class="pager__item pager__item--action pager__item--next">
          <a href="#" title="Go to next page" rel="next"
             aria-label="Next Page" class="next-page pager__link
             pager__link--action-link">Next</a>
        </li>
      `);

      // Add event listeners to pagination buttons.
      this.modalElement.find('.pagination .page-buttons a.number-page')
        .on('click', (event) => {
          event.preventDefault();
          this.currentPage = parseInt($(event.currentTarget).text());
          this.renderIcons();
        });

      // Add event listeners to "Previous" and "Next" buttons.
      this.modalElement.find('.pagination .prev-page')
        .on('click', this.prevPage.bind(this));
      this.modalElement.find('.pagination .next-page')
        .on('click', this.nextPage.bind(this));

      // Toggle "Previous" and "Next" buttons based on the current page state.
      this.modalElement.find('.pagination .prev-page')
        .toggle(this.currentPage !== 1);
      this.modalElement.find('.pagination .next-page')
        .toggle(this.currentPage !== this.totalPages);
    }

    /**
     * Updates the total icon count and page details in the UI.
     *
     * @param {number} totalIcons
     *   The total number of icons available after filtering.
     */
    updateStats(totalIcons) {
      this.statsElement.text(`${totalIcons} Icons`);
      this.currentPageElement.text(this.currentPage);
      this.totalPagesElement.text(this.totalPages);
    }

    /**
     * Updates counts and visibility of sidebar filters.
     *
     * This method updates sidebar filter counts, like styles and categories,
     * hiding filters with no matching icons dynamically.
     */
    updateSidebarCounts() {
      // Update the total icon count for "All styles".
      const totalIcons = this.filterIcons().length;
      this.sidebarTabs
        .find('[data-library-style="all"] .icons-facet-count')
        .text(totalIcons)
        .show();

      // Update counts and visibility for styles.
      const styles = this.getAllStyles();
      styles.forEach(style => {
        const styleCount = this.filterIconsBySidebar()
          .filter(icon => icon.style === style).length;

        const styleTab = this.sidebarTabs.find(`[data-library-style="${style}"]`);
        if (styleCount === 0) {
          styleTab.hide(); // Hide tabs with no matching icons.
        } else {
          styleTab.show(); // Show tabs with matching icons.
          if (this.settings.showCount) {
            styleTab
              .find('.icons-facet-count')
              .text(styleCount);
          }
        }
      });

      // Update counts and visibility for categories.
      Object.keys(this.source).forEach(category => {
        if (category !== 'Brands') {
          const categoryCount = this.filterIcons()
            .filter(icon =>
              Object.keys(this.source[category]).includes(icon.name) &&
              (this.activeFamily === 'all' || icon.family === this.activeFamily) &&
              (this.activeType === 'all' || icon.type === this.activeType)
            ).length;

          const categoryTab = this.sidebarTabs.find(`[data-library-id="${category}"]`);
          if (categoryCount === 0) {
            categoryTab.hide(); // Hide tabs with no matching icons.
          } else {
            categoryTab.show(); // Show tabs with matching icons.
            if (this.settings.showCount) {
              categoryTab
                .find('.icons-facet-count')
                .text(categoryCount);
            }
          }
        }
      });
    }

    /**
     * Filters icons based on active filters and search query.
     *
     * @return {Array}
     *   A filtered icon list based on sidebar selections and search input.
     */
    filterIconsBySidebar() {
      // Step 1: Aggregate all icons into a flat list with metadata.
      const uniqueIcons = new Map();
      Object.keys(this.source).forEach(category => {
        const categoryIcons = this.source[category];
        Object.keys(categoryIcons).forEach(iconName => {
          const iconData = categoryIcons[iconName];
          iconData.icons.forEach(icon => {
            const iconKey = `${iconName}-${icon.style}-${icon.family}`;
            if (!uniqueIcons.has(iconKey)) {
              uniqueIcons.set(iconKey, {
                ...icon,
                name: iconName,
                label: iconData.label,
                search_terms: iconData.search_terms,
                free: iconData.free,
                category: category,
              });
            }
          });
        });
      });

      // Convert unique icons to an array for filtering.
      let icons = Array.from(uniqueIcons.values());

      // Step 2: Filter by Family.
      if (this.activeFamily !== 'all') {
        const activeFamilies = this.modalElement
          .find('.icons-facet-tab.facet-active[data-family]')
          .map((_, elem) => $(elem).data('family'))
          .get();

        if (activeFamilies.length > 0) {
          icons = icons.filter(icon => activeFamilies.includes(icon.family));
        }
      }

      // Step 3: Filter by Style.
      if (this.activeStyle !== 'all') {
        const activeStyles = this.modalElement
          .find('.faui-modal--sidebar-tab-item.style-active[data-library-style]')
          .map((_, elem) => $(elem).data('library-style'))
          .get();

        if (activeStyles.length > 0) {
          icons = icons.filter(icon => activeStyles.includes(icon.style));
        }
      }

      // Step 4: Filter by Type (Free/Pro).
      if (this.activeType !== 'all') {
        icons = icons.filter(icon => {
          if (this.activeType === 'free') {
            // Include both classic and brand free icons.
            return (
              (icon.family === 'classic' && icon.free.includes(icon.style)) ||
              icon.style === 'brands'
            );
          }
          return icon.type === this.activeType; // Include only pro icons.
        });
      }

      // Step 5: Filter by Category.
      if (this.activeCategory !== 'all_categories') {
        icons = icons.filter(icon => icon.category === this.activeCategory);
      }

      // Step 6: Filter by Search Query.
      if (this.searchQuery) {
        const query = this.searchQuery.toLowerCase();
        icons = icons.filter(icon =>
          icon.search_terms.toLowerCase().includes(query)
        );
      }

      return icons; // Return the filtered list of icons.
    }

  }
})(jQuery);
